
#include <stdlib.h>
#include "assert.h"
#include "types.h"
#include "alpha.h"

static int b_has_one_arch(const board *b);
static int b_has_frame(const board *b);


static possible_move *
pmove(int x, int y, possible_move *next)
{
    possible_move *p = malloc(sizeof *p);
    assert(p);
    p->x = x; p->y = y;
    p->value = 0;
    p->next = next;
    return p;
}


possible_move *find_all_possible_moves(const board *b)
{
    possible_move *head = NULL;
    int x, y;

    for (y=0; y < b->h; ++y)
      for (x=0; x < b->w; ++x)
        if (is_present(b, x, y))
          head = pmove(x, y, head);

    return head;
}


/*
   Return a positive number if the board is good for player |b->turn|.
   We currently deal with the "one-arch" situation (loss).
   Todo fixme bug hack: Deal with "two-arch" situations (win).
   Deal with an "L" (win), "U" (loss), or "four sides" (win) pattern.
*/
int evaluate_ownership(const board *b)
{
    /* One arch left is a loss for the moving player. */
    if (b_has_one_arch(b)) return -1000;
    /* A "frame" is a loss for the moving player. */
    else if (b_has_frame(b)) return -1000;
    /* Otherwise, we have no idea. */
    else return 0;
}


static int
b_has_one_arch(const board *b)
{
    if (b->w < 2 || b->h < 2) return 0;
    if (b->npresent <= b->w && b->npresent <= b->h) return 1;
    if (b->npresent != b->w && b->npresent != b->h) return 0;

    /*
       At this point, we might have
       ..X..
       ..X..
       ..XXX
    */
    if (b->npresent == b->w) {
        int i, j;
        int found_row = 0;
        for (j=0; j < b->h; ++j) {
            for (i=0; i < b->w; ++i) {
                if (is_present(b, i, j)) {
                    if (found_row) goto failure1;
                    found_row = 1;
                    break;
                }
            }
        }
        assert(found_row == 1);
        return 1;
      failure1: ;
    }

    if (b->npresent == b->h) {
        int i, j;
        int found_col = 0;
        for (i=0; i < b->w; ++i) {
            for (j=0; j < b->h; ++j) {
                if (is_present(b, i, j)) {
                    if (found_col) goto failure2;
                    found_col = 1;
                    break;
                }
            }
        }
        assert(found_col == 1);
        return 1;
      failure2: ;
    }
    return 0;
}

static int
b_has_frame(const board *b)
{
    int i, j;

    if (b->w < 3 || b->h < 3) return 0;
    if (b->npresent != 2*(b->w+b->h) - 2) return 0;

    for (j=1; j < b->h-1; ++j) {
        if (!is_present(b, 0, j)) return 0;
        if (!is_present(b, b->w-1, j)) return 0;
        for (i=1; i < b->w-1; ++i) {
            if (is_present(b, i, j)) return 0;
        }
    }

    return 1;
}



/* Return the alpha-beta value of the given board for b->turn. */
possible_move *eval_alpha_beta_depth(const board *b, int depth)
{
    possible_move *poss, *best = NULL;
    poss = find_all_possible_moves(b);
    /* Note that having no moves means you win, not lose! */

    while (poss != NULL) {
        possible_move *cur = poss;
        board *b2 = copy_board(b);
        int this_e;
        knock_out(b2, cur->x, cur->y);
        if (depth > 0) {
            possible_move *m = eval_alpha_beta_depth(b2, depth-1);
            this_e = (m == NULL)? 1000: m->value;
            if (m) free(m);
        }
        else this_e = evaluate_ownership(b2);

        /* If we just placed one, then this_e refers to the other guy! */
        if (cur->x >= 0) this_e = -this_e;

        if (best == NULL || this_e > best->value) {
            cur->value = this_e;
            if (best) free(best);
            /* Don't free the new best move, then. */
            best = cur;
            cur = NULL;
        }
        kill_board(b2);
        poss = poss->next;
        free(cur);
        /* Did we find a winning move yet? */
        if (best && best->value == 1000) break;
    }
    /* The rest of the moves are irrelevant; we've found a winner. */
    while (poss != NULL) {
        possible_move *cur = poss;
        poss = poss->next;
        free(cur);
    }

    return best;
}


possible_move *eval_alpha_beta_moves(const board *b, int moves)
{
    /* Stub. Todo fixme bug hack. */
    (void)b; (void)moves;
    return NULL;
}
